r"""
Down-Up Algebras

AUTHORS:

- Travis Scrimshaw (2023-4): initial version
"""

# ****************************************************************************
#       Copyright (C) 2023 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.algebras import Algebras
from sage.categories.modules import Modules
from sage.categories.rings import Rings
from sage.combinat.free_module import CombinatorialFreeModule
from sage.sets.non_negative_integers import NonNegativeIntegers
from sage.categories.sets_cat import cartesian_product
from sage.sets.family import Family
from sage.misc.lazy_list import lazy_list
from sage.misc.misc_c import prod
from sage.modules.free_module import FreeModule


class DownUpAlgebra(CombinatorialFreeModule):
    r"""
    The down-up algebra.

    Let `R` be a commutative ring, and let `\alpha, \beta, \gamma \in R`.
    The *down-up algebra* is the associative unital algebra
    `DU(\alpha, \beta, \gamma)` generated by `d, u` with relations

    .. MATH::

        \begin{aligned}
        d^2u & = \alpha dud + \beta ud^2 + \gamma d,
        \\ du^2 & = \alpha udu + \beta u^2d + \gamma u.
        \end{aligned}

    The down-up algebra has a PBW-type basis given by

    .. MATH::

        \{ u^i (du)^j d^k \mid i,j,k \in \ZZ_{\geq 0} \}.

    This algebra originates in the study of posets. For a poset `P`,
    we define operators acting on `R[P]` by

    .. MATH::

        d(y) = \sum_x x \qquad\qquad u(y) = \sum_z z,

    where `y` covers `x` and `z` covers `y`. For `r`-differential posets
    we have `du - ud = r 1`, and thus it affords a representation of a
    :class:`Weyl algebra <sage.algebras.weyl_algebra.DifferentialWeylAlgebra>`.
    This Weyl algebra is obtained as the quotient of `DU(0, 1, 2r)` by the
    ideal generated by `du - ud - r`. For a `(q, r)`-differential poset,
    we have the `d` and `u` operators satisfying

    .. MATH::

        \begin{aligned}
        d^2u & = q(q+1) dud - q^3 ud^2 + r d,
        \\ du^2 & = q(q+1) udu - q^3 u^2d + r u,
        \end{aligned}

    or `\alpha = q(q+1)`, `\beta = -q^3`, and `\gamma = r`. Specializing
    `q = -1` recovers the `r`-differential poset relation.

    Two other noteworthy quotients are:

    - the `q`-Weyl algebra from `DU(0, q^2, q+1)` by the ideal generated by
      `du - qud - 1`, and
    - the quantum plane `R_q[d, u]`, where `du = qud`, from `DU(2q, -q^2, 0)`
      by the ideal generated by `du - qud`.

    EXAMPLES:

    We begin by constructing the down-up algebra and perform some
    basic computations::

        sage: R.<a,b,g> = QQ[]
        sage: DU = algebras.DownUp(a, b, g)
        sage: d, u = DU.gens()
        sage: d * u
        (d*u)
        sage: u * d
        u*d
        sage: d^2 * u
        b*u*d^2 + a*(d*u)*d + g*d
        sage: d * u^2
        b*u^2*d + a*u*(d*u) + g*u

    We verify some examples of Proposition 3.5 in [BR1998]_, which states
    that the `0`-th degree part is commutative::

        sage: DU0 = [u^i * (d*u)^j * d^i for i,j in
        ....:        cartesian_product([range(3), range(3)])]
        sage: all(x.degree() == 0 for x in DU0)
        True
        sage: all(x * y == y * x for x, y in cartesian_product([DU0, DU0]))
        True

    We verify that `DU(2, -1, \gamma)` can be described as the universal
    enveloping algebra of the 3-dimensional Lie algebra spanned by `x,y,z`
    satisfying `z = [x, y]`, `[x, z] = \gamma x`, and `[z, y] = \gamma y`::

        sage: R.<g> = QQ[]
        sage: L = LieAlgebra(R, {('x','y'): {'z': 1}, ('x','z'): {'x': g}, ('z','y'): {'y': g}},
        ....:                names='x,y,z')
        sage: x, y, z = L.basis()
        sage: (L[x, y], L[x, z], L[z, y])
        (z, g*x, g*y)
        sage: x, y, z = L.pbw_basis().gens()
        sage: x^2*y - 2*x*y*x + y*x^2 == g*x
        True
        sage: x*y^2 - 2*y*x*y + y^2*x == g*y
        True
        sage: DU = algebras.DownUp(2, -1, g)
        sage: d, u = DU.gens()
        sage: d^2*u - 2*d*u*d + u*d^2 == g*d
        True
        sage: d*u^2 - 2*u*d*u + u^2*d == g*u
        True

    Young's lattice is known to be a differential poset. Thus we can
    construct a representation of `DU(0, 1, 2)` on this poset (which
    gives a proof that Fomin's :class:`growth diagrams <GrowthDiagram>`
    are equivalent to edge local rules or shadow lines construction
    for :func:`RSK`)::

        sage: DU = algebras.DownUp(0, 1, 2)
        sage: d, u = DU.gens()
        sage: d^2*u == 0*d*u*d + 1*u*d*d + 2*d
        True
        sage: d*u^2 == 0*u*d*u + 1*u*u*d + 2*u
        True

        sage: YL = CombinatorialFreeModule(DU.base_ring(), Partitions())
        sage: def d_action(la):
        ....:     return YL.sum_of_monomials(la.remove_cell(*c) for c in la.removable_cells())
        sage: def u_action(la):
        ....:     return YL.sum_of_monomials(la.add_cell(*c) for c in la.addable_cells())
        sage: D = YL.module_morphism(on_basis=d_action, codomain=YL)
        sage: U = YL.module_morphism(on_basis=u_action, codomain=YL)
        sage: for la in PartitionsInBox(5, 5):
        ....:     b = YL.basis()[la]
        ....:     assert (D*D*U)(b) == 0*(D*U*D)(b) + 1*(U*D*D)(b) + 2*D(b)
        ....:     assert (D*U*U)(b) == 0*(U*D*U)(la) + 1*(U*U*D)(b) + 2*U(b)
        ....:     assert (D*U)(b) == (U*D)(b) + b  # the Weyl algebra relation

    .. TODO::

        Implement the homogenized version.

    REFERENCES:

    - [BR1998]_
    - [CM2000]_
    """
    @staticmethod
    def __classcall_private__(cls, alpha, beta, gamma, base_ring=None):
        r"""
        Standardize input to ensure a unique representation.

        TESTS::

            sage: R.<a,b,g> = QQ[]
            sage: DU1 = algebras.DownUp(a, 1, g)
            sage: DU2 = algebras.DownUp(a, R.one(), g)
            sage: DU3 = algebras.DownUp(a, 1, g, R)
            sage: DU1 is DU2 and DU2 is DU3
            True
        """
        if base_ring is None:
            from sage.structure.element import get_coercion_model
            base_ring = get_coercion_model().common_parent(alpha, beta, gamma)
        if base_ring not in Rings().Commutative():
            raise TypeError("base ring must be a commutative ring")
        alpha = base_ring(alpha)
        beta = base_ring(beta)
        gamma = base_ring(gamma)
        return super().__classcall__(cls, alpha, beta, gamma, base_ring=base_ring)

    def __init__(self, alpha, beta, gamma, base_ring):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: d, u = DU.gens()
            sage: elts = [d, u, d^2, u^2, d*u, u*d]
            sage: TestSuite(DU).run(elements=elts)
            sage: elts += [d*(d*u)*u]
            sage: TestSuite(DU).run(elements=elts)  # long time
        """
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma
        cat = Algebras(base_ring).WithBasis().Graded()
        if self._beta:
            from sage.categories.domains import Domains
            cat &= Domains()
        indices = cartesian_product([NonNegativeIntegers()] * 3)
        CombinatorialFreeModule.__init__(self, base_ring, indices, category=cat, sorting_reverse=True)
        self._assign_names(['d', 'u'])

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: DU = algebras.DownUp(1, 2, 3)
            sage: DU
            Down-Up algebra with parameters (1, 2, 3) over Integer Ring
        """
        return "Down-Up algebra with parameters ({}, {}, {}) over {}".format(
            self._alpha, self._beta, self._gamma, self.base_ring())

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: latex(DU)
            \mathcal{DU}(a,b,g)
        """
        return "\\mathcal{DU}(%s,%s,%s)" % (self._alpha, self._beta, self._gamma)

    def _repr_term(self, m) -> str:
        r"""
        Return a string representation of the basis element indexed by ``m``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: I = DU.indices()
            sage: DU._repr_term(I([1,0,5]))
            'u*d^5'
            sage: DU._repr_term(I([6,3,1]))
            'u^6*(d*u)^3*d'
            sage: DU._repr_term(I([0,1,2]))
            '(d*u)*d^2'
            sage: DU._repr_term(I([0,0,0]))
            '1'
        """
        if not any(m):
            return '1'
        ret = ''
        for i, s in enumerate(['u', '(d*u)', 'd']):
            if not m[i]:
                continue
            if ret:
                ret += '*'
            if m[i] == 1:
                ret += s
            else:
                ret += f"{s}^{m[i]}"
        return ret

    def _latex_term(self, m) -> str:
        r"""
        Return a latex representation for the basis element indexed by ``m``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: I = DU.indices()
            sage: DU._latex_term(I([1,0,5]))
            'ud^{5}'
            sage: DU._latex_term(I([6,3,1]))
            'u^{6}(du)^{3}d'
            sage: DU._latex_term(I([0,1,2]))
            '(du)d^{2}'
            sage: DU._latex_term(I([0,0,0]))
            '1'
        """
        if all(val == 0 for val in m):
            return '1'
        ret = ''
        for i, s in enumerate(['u', '(du)', 'd']):
            if not m[i]:
                continue
            if m[i] == 1:
                ret += s
            else:
                ret += f"{s}^{{{m[i]}}}"
        return ret

    @cached_method
    def algebra_generators(self):
        r"""
        Return the algebra generators of ``self``.

        EXAMPLES::

            sage: DU = algebras.DownUp(2, 3, 4)
            sage: dict(DU.algebra_generators())
            {'d': d, 'u': u}
        """
        u = self.monomial(self._indices([1,0,0]))
        d = self.monomial(self._indices([0,0,1]))
        return Family({'d': d, 'u': u})

    @cached_method
    def gens(self) -> tuple:
        r"""
        Return the generators of ``self``.

        EXAMPLES::

            sage: DU = algebras.DownUp(2, 3, 4)
            sage: DU.gens()
            (d, u)
        """
        G = self.algebra_generators()
        return (G['d'], G['u'])

    @cached_method
    def one_basis(self):
        r"""
        Return the index of the basis element of `1`.

        EXAMPLES::

            sage: DU = algebras.DownUp(2, 3, 4)
            sage: DU.one_basis()
            (0, 0, 0)
        """
        return self._indices([0, 0, 0])

    def product_on_basis(self, m1, m2):
        r"""
        Return the product of the basis elements indexed by ``m1`` and ``m2``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: I = DU.indices()
            sage: DU.product_on_basis(I([2,0,0]), I([4,0,0]))
            u^6
            sage: DU.product_on_basis(I([2,0,0]), I([0,4,0]))
            u^2*(d*u)^4
            sage: DU.product_on_basis(I([2,0,0]), I([0,0,4]))
            u^2*d^4
            sage: DU.product_on_basis(I([0,2,0]), I([0,4,0]))
            (d*u)^6
            sage: DU.product_on_basis(I([0,2,0]), I([0,0,4]))
            (d*u)^2*d^4
            sage: DU.product_on_basis(I([0,0,2]), I([0,0,4]))
            d^6
            sage: DU.product_on_basis(I([5,3,1]), I([1,0,4]))
            u^5*(d*u)^4*d^4

            sage: DU.product_on_basis(I([0,1,0]), I([1,0,0]))
            b*u^2*d + a*u*(d*u) + g*u
            sage: DU.product_on_basis(I([0,0,2]), I([1,0,0]))
            b*u*d^2 + a*(d*u)*d + g*d
            sage: DU.product_on_basis(I([0,0,1]), I([2,0,0]))
            b*u^2*d + a*u*(d*u) + g*u
            sage: DU.product_on_basis(I([0,0,1]), I([0,1,0]))
            b*u*d^2 + a*(d*u)*d + g*d

            sage: DU.product_on_basis(I([0,1,0]), I([3,0,0]))
            (a^2*b+b^2)*u^4*d + (a^3+2*a*b)*u^3*(d*u) + (a^2*g+a*g+b*g+g)*u^3
            sage: DU.product_on_basis(I([1,1,3]), I([0,1,1]))
            (a^2*b^2+b^3)*u^3*d^6 + (a^3*b+a*b^2)*u^2*(d*u)*d^5 + (a^2*b*g+b^2*g)*u^2*d^5
             + (a^3+2*a*b)*u*(d*u)^2*d^4 + (a^2*g+a*g+b*g+g)*u*(d*u)*d^4
        """
        # Check trivial cases
        if not any(m1):
            return self.monomial(m2)
        if not any(m2):
            return self.monomial(m1)

        u1, du1, d1 = m1
        u2, du2, d2 = m2
        I = self._indices

        if not d1:
            if not u2:
                return self.monomial(I([u1, du1+du2, d2]))
            # else u2 > 0
            if not du1:
                return self.monomial(I([u1+u2, du2, d2]))
            # Perform du * u reduction
            lhs = self.monomial(I([u1, du1-1, 0]))
            mid = self._from_dict({I([1,1,0]): self._alpha,
                                   I([2,0,1]): self._beta,
                                   I([1,0,0]): self._gamma})
            rhs = self.monomial(I([u2-1, du2, d2]))
        else:  # d1 > 0
            if not u2:
                if not du2:
                    return self.monomial(I([u1, du1, d1+d2]))
                # Perform a d * du reduction
                lhs = self.monomial(I([u1, du1, d1-1]))
                mid = self._from_dict({I([0,1,1]): self._alpha,
                                       I([1,0,2]): self._beta,
                                       I([0,0,1]): self._gamma})
                rhs = self.monomial(I([0, du2-1, d2]))
            elif u2 > 1:
                # Perform d * u^2 reduction
                lhs = self.monomial(I([u1, du1, d1-1]))
                mid = self._from_dict({I([1,1,0]): self._alpha,
                                       I([2,0,1]): self._beta,
                                       I([1,0,0]): self._gamma})
                rhs = self.monomial(I([u2-2, du2, d2]))
            elif u2 == 1:
                if d1 == 1:
                    return self.monomial(I([u1, du1+du2+1, d2]))
                # Perform a d^2 * u reduction
                lhs = self.monomial(I([u1, du1, d1-2]))
                mid = self._from_dict({I([0,1,1]): self._alpha,
                                       I([1,0,2]): self._beta,
                                       I([0,0,1]): self._gamma})
                rhs = self.monomial(I([0, du2, d2]))

        if lhs == self.one():
            if rhs == self.one():
                return mid
            return mid * rhs
        if rhs == self.one():
            return lhs * mid
        return lhs * mid * rhs

    def degree_on_basis(self, m):
        r"""
        Return the degree of the basis element indexed by ``m``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: I = DU.indices()
            sage: DU.degree_on_basis(I([0, 3, 2]))
            -2
            sage: DU.degree_on_basis(I([2, 3, 0]))
            2
            sage: DU.degree_on_basis(I([2, 0, 3]))
            -1
            sage: DU.degree_on_basis(I([3, 10, 3]))
            0
        """
        return m[0] - m[2]

    def verma_module(self, la):
        r"""
        Return the :class:`Verma module
        <sage.algebras.down_up_algebra.VermaModule>`
        `V(\lambda)` of ``self``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: DU.verma_module(5)
            Verma module of weight 5 of Down-Up algebra with parameters (a, b, g)
             over Multivariate Polynomial Ring in a, b, g over Rational Field
        """
        return VermaModule(self, la)


class VermaModule(CombinatorialFreeModule):
    r"""
    The Verma module `V(\lambda)` of a down-up algebra.

    The Verma module `V(\lambda)` for the down-up algebra generated
    by `d, u` is the span of `\{v_n \mid n \in \ZZ_{\geq 0} \}`
    satisfying the relations

    .. MATH::

        d \cdot v_n = \lambda_{n-1} v_{n-1}, \qquad\qquad
        u \cdot v_n = v_{n+1},

    where `\lambda_n = \alpha \lambda_{n-1} + \beta \lambda_{n-2} + \gamma`
    and we set `\lambda_0 = \lambda` and `\lambda_{-1} = 0`.

    By Proposition 2.4 in [BR1998]_, `V(\lambda)` is simple if and
    only if `\lambda_n \neq 0` for all `n \geq 0`. Moreover, a maximal
    submodule is spanned by `\{ v_n \mid n > m \}`, where `m` is the
    minimal index such that `\lambda_m = 0`. Moreover, this is unique
    unless `\gamma = \lambda = 0`.

    EXAMPLES::

        sage: R.<a,b> = QQ[]
        sage: DU = algebras.DownUp(0, b, 1)
        sage: d, u = DU.gens()
        sage: V = DU.verma_module(a)
        sage: list(V.weights()[:6])
        [a, 1, a*b + 1, b + 1, a*b^2 + b + 1, b^2 + b + 1]
        sage: v = V.basis()
        sage: d^2 * v[2]
        a*v[0]
        sage: d * (d * v[2])
        a*v[0]

    The weight is computed by looking at the scalars associated to the
    action of `du` and `ud`::

        sage: d*u * v[3]
        (b+1)*v[3]
        sage: u*d * v[3]
        (a*b+1)*v[3]
        sage: v[3].weight()
        (b + 1, a*b + 1)

    An `U(\mathfrak{sl}_2)` example::

        sage: DU = algebras.DownUp(2, -1, -2)
        sage: d, u = DU.gens()
        sage: V = DU.verma_module(5)
        sage: list(V.weights()[:10])
        [5, 8, 9, 8, 5, 0, -7, -16, -27, -40]
        sage: v6 = V.basis()[6]
        sage: d * v6
        0
        sage: [V.basis()[i].weight() for i in range(6)]
        [(5, 0), (8, 5), (9, 8), (8, 9), (5, 8), (0, 5)]

    Note that these are the same `\mathfrak{sl}_2` weights from the usual
    construction of the irreducible representation `V(5)` (but they are
    different as `\mathfrak{gl}_2` weights)::

        sage: B = crystals.Tableaux(['A',1], shape=[5])
        sage: [b.weight() for b in B]
        [(5, 0), (4, 1), (3, 2), (2, 3), (1, 4), (0, 5)]

    An example with periodic weights (see Theorem 2.13 of [BR1998]_)::

        sage: k.<z6> = CyclotomicField(6)
        sage: al = z6 + 1
        sage: (al - 1)^6 == 1
        True
        sage: DU = algebras.DownUp(al, 1-al, 0)
        sage: V = DU.verma_module(5)
        sage: list(V.weights()[:8])
        [5, 5*z6 + 5, 10*z6, 10*z6 - 5, 5*z6 - 5, 0, 5, 5*z6 + 5]
    """
    @staticmethod
    def __classcall_private__(cls, DU, la):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: from sage.algebras.down_up_algebra import VermaModule
            sage: VermaModule(DU, 5) is VermaModule(DU, R(5))
            True
            sage: VermaModule(DU, 1/a)
            Traceback (most recent call last):
            ...
            TypeError: fraction must have unit denominator
        """
        R = DU.base_ring()
        la = R(la)
        return super().__classcall__(cls, DU, la)

    def __init__(self, DU, la):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: V = DU.verma_module(5)
            sage: TestSuite(V).run()
            sage: V = DU.verma_module(0)
            sage: TestSuite(V).run()

            sage: DU = algebras.DownUp(a, 0, g)
            sage: V = DU.verma_module(5)
            sage: TestSuite(V).run()
            sage: V = DU.verma_module(0)
            sage: TestSuite(V).run()

            sage: DU = algebras.DownUp(a, 1-a, 0)
            sage: V = DU.verma_module(5)
            sage: TestSuite(V).run()
            sage: V = DU.verma_module(0)
            sage: TestSuite(V).run()
        """
        self._DU = DU
        R = DU.base_ring()

        def _la_iter():
            m2 = la
            yield la
            m2 = R.zero()
            m1 = la
            while True:
                cur = DU._alpha * m1 + DU._beta * m2 + DU._gamma
                yield cur
                m2 = m1
                m1 = cur

        self._weights = lazy_list(_la_iter())
        cat = Modules(R).WithBasis()
        CombinatorialFreeModule.__init__(self, R, NonNegativeIntegers(),
                                         prefix='v', category=cat)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: DU = algebras.DownUp(1, 2, 3)
            sage: DU.verma_module(5)
            Verma module of weight 5 of Down-Up algebra with parameters (1, 2, 3) over Integer Ring
        """
        return f"Verma module of weight {self._weights[0]} of {self._DU}"

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: DU = algebras.DownUp(1, 2, 3)
            sage: latex(DU.verma_module(5))
            V\left(5\right)
        """
        return f"V\\left({self._weights[0]}\\right)"

    def highest_weight_vector(self):
        r"""
        Return the highest weight vector of ``self`` that generates
        ``self`` as a down-up module.

        EXAMPLES::

            sage: DU = algebras.DownUp(1, 2, 3)
            sage: V = DU.verma_module(5)
            sage: V.highest_weight_vector()
            v[0]
        """
        return self.basis()[0]

    def weights(self):
        r"""
        Return the sequence of weights `(\lambda_n)_{n=0}^{\infty}`.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(a, b, g)
            sage: V = DU.verma_module(5)
            sage: V.weights()
            lazy list [5, 5*a + g, 5*a^2 + a*g + 5*b + g, ...]

            sage: V = DU.verma_module(0)
            sage: DU = algebras.DownUp(a, 1-a, 0)
            sage: V = DU.verma_module(0)
            sage: V.weights()
            lazy list [0, 0, 0, ...]

        We reproduce the Fibonacci numbers example from [BR1998]_::

            sage: R.<la> = QQ[]
            sage: DU = algebras.DownUp(1, 1, 0, R)
            sage: V = DU.verma_module(la)
            sage: list(V.weights()[:11])
            [la, la, 2*la, 3*la, 5*la, 8*la, 13*la, 21*la, 34*la, 55*la, 89*la]
        """
        return self._weights

    def _action_on_basis(self, m, n):
        r"""
        Return the action of a basis element of the down-up algebra indexed
        by ``m`` on the basis element of ``self`` indexed by ``n``.

        EXAMPLES::

            sage: R.<a,b,g> = QQ[]
            sage: DU = algebras.DownUp(0, b, 1)
            sage: I = DU.indices()
            sage: V = DU.verma_module(1)
            sage: V.weights()
            lazy list [1, 1, b + 1, ...]
            sage: V._action_on_basis(I([0,0,1]), 0)
            0
            sage: V._action_on_basis(I([0,1,0]), 0)
            v[0]
            sage: V._action_on_basis(I([1,0,0]), 0)
            v[1]

            sage: V._action_on_basis(I([0,0,1]), 3)
            (b+1)*v[2]
            sage: V._action_on_basis(I([0,1,0]), 3)
            (b+1)*v[3]
            sage: V._action_on_basis(I([1,0,0]), 3)
            v[4]

            sage: V._action_on_basis(I([0,0,3]), 3)
            (b+1)*v[0]
            sage: V._action_on_basis(I([1,2,1]), 3)
            (b^3+3*b^2+3*b+1)*v[3]

            sage: V = DU.verma_module(0)
            sage: V._action_on_basis(I([0,0,1]), 1)
            0
        """
        if m[2] > n:
            return self.zero()
        np = n - m[2]
        coeff = prod(self._weights[np:n]) * self._weights[np] ** m[1]
        return self.term(n - m[2] + m[0], coeff)

    class Element(CombinatorialFreeModule.Element):
        r"""
        An element of a Verma module of a down-up algebra.
        """
        def _acted_upon_(self, scalar, self_on_left):
            r"""
            Return the action of ``scalar`` (an element of the base ring or
            the defining down-up algebra) on ``self``.

            EXAMPLES::

                sage: R.<a,b> = QQ[]
                sage: DU = algebras.DownUp(0, b, 1)
                sage: d, u = DU.gens()
                sage: V = DU.verma_module(a)
                sage: it = iter(DU.basis())
                sage: scalars = [next(it) for _ in range(10)]; scalars
                [1, u, (d*u), d, u^2, u*(d*u), u*d, (d*u)^2, (d*u)*d, d^2]
                sage: vecs = [V.basis()[0], V.basis()[1], V.basis()[6]]
                sage: all((x * y) * v == x * (y * v)
                ....:     for x in scalars for y in scalars for v in vecs)
                True
                sage: 5 * V.basis()[3]
                5*v[3]
                sage: V.basis()[0] * d
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *:
                 'Verma module of weight a of Down-Up algebra ...'
                 and 'Down-Up algebra ...'
            """
            ret = super()._acted_upon_(scalar, self_on_left)
            if ret is not None:
                return ret
            P = self.parent()
            try:
                scalar = P._DU(scalar)
            except (TypeError, ValueError):
                return None
            if self_on_left:
                return None
            return P.linear_combination((P._action_on_basis(m, n), mc*nc)
                                        for m, mc in scalar._monomial_coefficients.items()
                                        for n, nc in self._monomial_coefficients.items())

        def is_weight_vector(self) -> bool:
            r"""
            Return whether ``self`` is a weight vector.

            EXAMPLES::

                sage: DU = algebras.DownUp(2, -1, -2)
                sage: V = DU.verma_module(5)
                sage: V.zero().is_weight_vector()
                False
                sage: B = V.basis()
                sage: [B[i].weight() for i in range(6)]
                [(5, 0), (8, 5), (9, 8), (8, 9), (5, 8), (0, 5)]
                sage: B[5].is_weight_vector()
                True
                sage: v = B[0] + B[1]
                sage: v.is_weight_vector()
                False

                sage: DU = algebras.DownUp(2, -1, 0)
                sage: V = DU.verma_module(0)
                sage: B = V.basis()
                sage: v = sum(i*B[i] for i in range(1,5))
                sage: v.is_weight_vector()
                True
            """
            if not self:
                return False

            P = self.parent()
            R = P.base_ring()

            def get_wt(n):
                if not n:
                    return (R(P._weights[0]), R.zero())
                return (R(P._weights[n]), R(P._weights[n-1]))

            it = iter(self._monomial_coefficients)
            wt = get_wt(next(it))
            return all(get_wt(n) == wt for n in it)

        def weight(self):
            r"""
            Return the weight of ``self``.

            For `v_n`, this is the vector with the pair
            `(\lambda_n, \lambda_{n-1})`.

            EXAMPLES::

                sage: R.<a,b,g> = QQ[]
                sage: DU = algebras.DownUp(a, b, g)
                sage: V = DU.verma_module(5)
                sage: B = V.basis()
                sage: B[0].weight()
                (5, 0)
                sage: B[1].weight()
                (5*a + g, 5)
                sage: B[2].weight()
                (5*a^2 + a*g + 5*b + g, 5*a + g)

                sage: V.zero().weight()
                Traceback (most recent call last):
                ...
                ValueError: the zero element does not have well-defined weight
                sage: (B[0] + B[1]).weight()
                Traceback (most recent call last):
                ...
                ValueError: not a weight vector
            """
            if not self:
                raise ValueError("the zero element does not have well-defined weight")
            if not self.is_weight_vector():
                raise ValueError("not a weight vector")
            P = self.parent()
            R = P.base_ring()
            V = FreeModule(R, 2)
            it = iter(self._monomial_coefficients)
            n = next(it)
            if not n:
                return V([P._weights[0], R.zero()])
            return V([P._weights[n], P._weights[n - 1]])



# sage.doctest: optional - scipy
"""
Numerical Root Finding and Optimization

AUTHOR:

- William Stein (2007): initial version
- Nathann Cohen (2008): Bin Packing


Functions and Methods
----------------------
"""

from sage.modules.free_module_element import vector
from sage.rings.real_double import RDF


def find_root(f, a, b, xtol=10e-13, rtol=2.0**-50, maxiter=100, full_output=False):
    r"""
    Numerically find a root of ``f`` on the closed interval `[a,b]`
    (or `[b,a]`) if possible, where ``f`` is a function in the one variable.
    Note: this function only works in fixed (machine) precision, it is not
    possible to get arbitrary precision approximations with it.

    INPUT:

    - ``f`` -- a function of one variable or symbolic equality

    - ``a``, ``b`` -- endpoints of the interval

    - ``xtol``, ``rtol`` -- the routine converges when a root is known
      to lie within ``xtol`` of the value return. Should be `\geq 0`.
      The routine modifies this to take into account the relative precision
      of doubles. By default, rtol is ``4*numpy.finfo(float).eps``, the
      minimum allowed value for :func:`scipy:scipy.optimize.brentq`, which is
      what this method uses underneath. This value is equal to ``2.0**-50`` for
      IEEE-754 double precision floats as used by Python.

    - ``maxiter`` -- integer; if convergence is not achieved in
      ``maxiter`` iterations, an error is raised. Must be `\geq 0`.

    - ``full_output`` -- boolean (default: ``False``); if ``True``, also return
      object that contains information about convergence

    EXAMPLES:

    An example involving an algebraic polynomial function::

        sage: R.<x> = QQ[]
        sage: f = (x+17)*(x-3)*(x-1/8)^3
        sage: find_root(f, 0,4)
        2.999999999999995
        sage: find_root(f, 0,1)  # abs tol 1e-6 (note -- precision of answer isn't very good on some machines)
        0.124999
        sage: find_root(f, -20,-10)
        -17.0

    In Pomerance's book on primes he asserts that the famous Riemann
    Hypothesis is equivalent to the statement that the function `f(x)`
    defined below is positive for all `x \geq 2.01`::

        sage: def f(x):
        ....:     return sqrt(x) * log(x) - abs(Li(x) - prime_pi(x))

    We find where `f` equals, i.e., what value that is slightly smaller
    than `2.01` that could have been used in the formulation of the Riemann
    Hypothesis::

        sage: find_root(f, 2, 4, rtol=0.0001)
        2.0082...

    This agrees with the plot::

        sage: plot(f,2,2.01)
        Graphics object consisting of 1 graphics primitive

    The following example was added due to :issue:`4942` and demonstrates that
    the function need not be defined at the endpoints::

        sage: find_root(x^2*log(x,2)-1,0, 2)  # abs tol 1e-6
        1.41421356237

    The following is an example, again from :issue:`4942` where Brent's method
    fails. Currently no other method is implemented, but at least we
    acknowledge the fact that the algorithm fails::

        sage: find_root(1/(x-1)+1,0, 2)
        0.0
        sage: find_root(1/(x-1)+1,0.00001, 2)
        Traceback (most recent call last):
        ...
        NotImplementedError: Brent's method failed to find a zero for f on the interval

    An example of a function which evaluates to NaN on the entire interval::

        sage: f(x) = 0.0 / max(0, x)
        sage: find_root(f, -1, 0)
        Traceback (most recent call last):
        ...
        RuntimeError: f appears to have no zero on the interval
    """
    try:
        return f.find_root(a=a,b=b,xtol=xtol,rtol=rtol,maxiter=maxiter,full_output=full_output)
    except AttributeError:
        pass
    a = float(a)
    b = float(b)
    if a > b:
        a, b = b, a
    left = f(a)
    right = f(b)

    if left > 0 and right > 0:
        # Refine further -- try to find a point where this
        # function is negative in the interval
        val, s = find_local_minimum(f, a, b)
        if val > 0:
            if val < rtol:
                if full_output:
                    return s, "No extra data"
                return s
            raise RuntimeError("f appears to have no zero on the interval")
        # If we found such an s, then we just instead find
        # a root between left and s or s and right.
        a = s   # arbitrary choice -- maybe should try both and take one that works?

    elif left < 0 and right < 0:
        # Refine further
        val, s = find_local_maximum(f, a, b)
        if val < 0:
            if abs(val) < rtol:
                if full_output:
                    return s, "No extra data"
                return s
            raise RuntimeError("f appears to have no zero on the interval")
        a = s

    # Fixing :issue:`4942` - if the answer on any of the endpoints is NaN,
    # we restrict to looking between minimum and maximum values in the segment
    # Note - this could be used in all cases, but it requires some more
    # computation

    if (left != left) or (right != right):
        minval, s_1 = find_local_minimum(f, a, b)
        maxval, s_2 = find_local_maximum(f, a, b)
        if ((minval > 0) or (maxval < 0) or
           (minval != minval) or (maxval != maxval)):
            raise RuntimeError("f appears to have no zero on the interval")
        a = min(s_1, s_2)
        b = max(s_1, s_2)

    import scipy.optimize
    import numpy
    if int(numpy.version.short_version[0]) > 1:
        numpy.set_printoptions(legacy="1.25")

    g = lambda x: float(f(x))
    brentqRes = scipy.optimize.brentq(g, a, b,
                                 full_output=full_output, xtol=xtol, rtol=rtol, maxiter=maxiter)
    # A check following :issue:`4942`, to ensure we actually found a root
    # Maybe should use a different tolerance here?
    # The idea is to take roughly the derivative and multiply by estimated
    # value of the root
    root = 0
    if full_output:
        root = brentqRes[0]
    else:
        root = brentqRes
    if abs(f(root)) > max(abs(root * rtol * (right - left) / (b - a)), 1e-6):
        raise NotImplementedError("Brent's method failed to find a zero for f on the interval")
    return brentqRes


def find_local_maximum(f, a, b, tol=1.48e-08, maxfun=500):
    """
    Numerically find a local maximum of the expression `f` on the interval
    `[a,b]` (or `[b,a]`) along with the point at which the maximum is attained.

    Note that this function only finds a *local* maximum, and not the
    global maximum on that interval -- see the examples with
    :func:`find_local_maximum`.

    See the documentation for :func:`find_local_maximum` for more
    details and possible workarounds for finding the global minimum on
    an interval.

    EXAMPLES::

        sage: f = lambda x: x*cos(x)
        sage: find_local_maximum(f, 0, 5)
        (0.561096338191..., 0.8603335890...)
        sage: find_local_maximum(f, 0, 5, tol=0.1, maxfun=10)
        (0.561090323458..., 0.857926501456...)
        sage: find_local_maximum(8*e^(-x)*sin(x) - 1, 0, 7)                             # needs sage.symbolic
        (1.579175535558..., 0.7853981...)
    """
    try:
        return f.find_local_maximum(a=a, b=b, tol=tol, maxfun=maxfun)
    except AttributeError:
        pass
    minval, x = find_local_minimum(lambda z: -f(z), a=a, b=b, tol=tol, maxfun=maxfun)
    return -minval, x


def find_local_minimum(f, a, b, tol=1.48e-08, maxfun=500):
    """
    Numerically find a local minimum of the expression ``f`` on the
    interval `[a,b]` (or `[b,a]`) and the point at which it attains that
    minimum.  Note that ``f`` must be a function of (at most) one
    variable.

    Note that this function only finds a *local* minimum, and not the
    global minimum on that interval -- see the examples below.

    INPUT:

    - ``f`` -- a function of at most one variable

    - ``a``, ``b`` -- endpoints of interval on which to minimize `f`

    - ``tol`` -- the convergence tolerance

    - ``maxfun`` -- maximum function evaluations

    OUTPUT:

    - ``minval`` -- (float) the minimum value that `f` takes on in the
      interval `[a,b]`

    - ``x`` -- (float) the point at which `f` takes on the minimum value

    EXAMPLES::

        sage: f = lambda x: x*cos(x)
        sage: find_local_minimum(f, 1, 5)
        (-3.28837139559..., 3.4256184695...)
        sage: find_local_minimum(f, 1, 5, tol=1e-3)
        (-3.28837136189098..., 3.42575079030572...)
        sage: find_local_minimum(f, 1, 5, tol=1e-2, maxfun=10)
        (-3.28837084598..., 3.4250840220...)
        sage: show(plot(f, 0, 20))                                                      # needs sage.plot
        sage: find_local_minimum(f, 1, 15)
        (-9.4772942594..., 9.5293344109...)

    Only local minima are found; if you enlarge the interval, the
    returned minimum may be *larger*! See :issue:`2607`.

    ::

        sage: # needs sage.symbolic
        sage: f(x) = -x*sin(x^2)
        sage: find_local_minimum(f, -2.5, -1)
        (-2.182769784677722, -2.1945027498534686)

    Enlarging the interval returns a larger minimum::

        sage: # needs sage.symbolic
        sage: find_local_minimum(f, -2.5, 2)
        (-1.3076194129914434, 1.3552111405712108)

    One work-around is to plot the function and grab the minimum from
    that, although the plotting code does not necessarily do careful
    numerics (observe the small number of decimal places that we
    actually test)::

        sage: # needs sage.plot sage.symbolic
        sage: plot(f, (x, -2.5, -1)).ymin()
        -2.182...
        sage: plot(f, (x, -2.5, 2)).ymin()
        -2.182...

    ALGORITHM:

    Uses :func:`scipy:scipy.optimize.fminbound` which uses Brent's method.


    AUTHOR:

    - William Stein (2007-12-07)
    """
    try:
        return f.find_local_minimum(a=a, b=b, tol=tol, maxfun=maxfun)
    except AttributeError:
        pass
    a = float(a)
    b = float(b)
    import scipy.optimize
    import numpy
    if int(numpy.version.short_version[0]) > 1:
        numpy.set_printoptions(legacy="1.25")

    xmin, fval, iter, funcalls = scipy.optimize.fminbound(f, a, b, full_output=1, xtol=tol, maxfun=maxfun)
    return fval, xmin


def minimize(func, x0, gradient=None, hessian=None, algorithm='default',
             verbose=False, **args):
    r"""
    This function is an interface to a variety of algorithms for computing
    the minimum of a function of several variables.

    INPUT:

    - ``func`` -- either a symbolic function or a Python function whose
      argument is a tuple with `n` components

    - ``x0`` -- initial point for finding minimum

    - ``gradient`` -- (optional) gradient function. This will be computed
      automatically for symbolic functions.  For Python functions, it allows
      the use of algorithms requiring derivatives.  It should accept a
      tuple of arguments and return a NumPy array containing the partial
      derivatives at that point.

    - ``hessian`` -- (optional) hessian function. This will be computed
      automatically for symbolic functions. For Python functions, it allows
      the use of algorithms requiring derivatives. It should accept a tuple
      of arguments and return a NumPy array containing the second partial
      derivatives of the function.

    - ``algorithm`` -- string specifying algorithm to use. Options are
      ``'default'`` (for Python functions, the simplex method is the default)
      (for symbolic functions bfgs is the default):

      - ``'simplex'`` -- using the downhill simplex algorithm

      - ``'powell'`` -- use the modified Powell algorithm

      - ``'bfgs'`` -- (Broyden-Fletcher-Goldfarb-Shanno) requires gradient

      - ``'cg'`` -- (conjugate-gradient) requires gradient

      - ``'ncg'`` -- (newton-conjugate gradient) requires gradient and hessian

    - ``verbose`` -- boolean (default: ``False``); print convergence message

    .. NOTE::

        For additional information on the algorithms implemented in this function,
        consult SciPy's :mod:`documentation on optimization and root
        finding <scipy:scipy.optimize>`.

    EXAMPLES:

    Minimize a fourth order polynomial in three variables (see the
    :wikipedia:`Rosenbrock_function`)::

        sage: vars = var('x y z')                                                       # needs sage.symbolic
        sage: f = 100*(y-x^2)^2 + (1-x)^2 + 100*(z-y^2)^2 + (1-y)^2                     # needs sage.symbolic
        sage: minimize(f, [.1,.3,.4])  # abs tol 1e-6                                   # needs sage.symbolic
        (1.0, 1.0, 1.0)

    Try the newton-conjugate gradient method; the gradient and hessian are
    computed automatically::

        sage: minimize(f, [.1, .3, .4], algorithm='ncg')  # abs tol 1e-6                # needs sage.symbolic
        (1.0, 1.0, 1.0)

    We get additional convergence information with the `verbose` option::

        sage: minimize(f, [.1, .3, .4], algorithm='ncg', verbose=True)                  # needs sage.symbolic
        Optimization terminated successfully.
        ...
        (0.9999999..., 0.999999..., 0.999999...)

    Same example with just Python functions::

        sage: def rosen(x):  # The Rosenbrock function
        ....:    return sum(100.0r*(x[1r:]-x[:-1r]**2.0r)**2.0r + (1r-x[:-1r])**2.0r)
        sage: minimize(rosen, [.1,.3,.4]) # abs tol 3e-5
        (1.0, 1.0, 1.0)

    Same example with a pure Python function and a Python function to
    compute the gradient::

        sage: # needs numpy
        sage: def rosen(x):  # The Rosenbrock function
        ....:    return sum(100.0r*(x[1r:]-x[:-1r]**2.0r)**2.0r + (1r-x[:-1r])**2.0r)
        sage: import numpy
        sage: if int(numpy.version.short_version[0]) > 1:
        ....:     _ = numpy.set_printoptions(legacy="1.25")
        sage: from numpy import zeros
        sage: def rosen_der(x):
        ....:    xm = x[1r:-1r]
        ....:    xm_m1 = x[:-2r]
        ....:    xm_p1 = x[2r:]
        ....:    der = zeros(x.shape, dtype=float)
        ....:    der[1r:-1r] = 200r*(xm-xm_m1**2r) - 400r*(xm_p1 - xm**2r)*xm - 2r*(1r-xm)
        ....:    der[0] = -400r*x[0r]*(x[1r]-x[0r]**2r) - 2r*(1r-x[0])
        ....:    der[-1] = 200r*(x[-1r]-x[-2r]**2r)
        ....:    return der
        sage: minimize(rosen, [.1,.3,.4], gradient=rosen_der,  # abs tol 1e-6
        ....:          algorithm='bfgs')
        (1.0, 1.0, 1.0)
    """
    from sage.structure.element import Expression
    from sage.ext.fast_callable import fast_callable
    import numpy
    if int(numpy.version.short_version[0]) > 1:
        numpy.set_printoptions(legacy="1.25")

    from scipy import optimize
    if isinstance(func, Expression):
        var_list = func.variables()
        var_names = [str(_) for _ in var_list]
        fast_f = fast_callable(func, vars=var_names, domain=float)
        f = lambda p: fast_f(*p)
        gradient_list = func.gradient()
        fast_gradient_functions = [fast_callable(gradient_list[i],
                                                 vars=var_names, domain=float)
                                   for i in range(len(gradient_list))]
        gradient = lambda p: numpy.array([ a(*p) for a in fast_gradient_functions])
    else:
        f = func

    if algorithm == "default":
        if gradient is None:
            min = optimize.fmin(f, [float(_) for _ in x0], disp=verbose, **args)
        else:
            min = optimize.fmin_bfgs(f, [float(_) for _ in x0],fprime=gradient, disp=verbose, **args)
    else:
        if algorithm == "simplex":
            min = optimize.fmin(f, [float(_) for _ in x0], disp=verbose, **args)
        elif algorithm == "bfgs":
            min = optimize.fmin_bfgs(f, [float(_) for _ in x0], fprime=gradient, disp=verbose, **args)
        elif algorithm == "cg":
            min = optimize.fmin_cg(f, [float(_) for _ in x0], fprime=gradient, disp=verbose, **args)
        elif algorithm == "powell":
            min = optimize.fmin_powell(f, [float(_) for _ in x0], disp=verbose, **args)
        elif algorithm == "ncg":
            if isinstance(func, Expression):
                hess = func.hessian()
                hess_fast = [ [fast_callable(a, vars=var_names, domain=float) for a in row] for row in hess]
                hessian = lambda p: [[a(*p) for a in row] for row in hess_fast]
                from numpy import dot
                hessian_p = lambda p,v: dot(numpy.array(hessian(p)),v)
                min = optimize.fmin_ncg(f, [float(_) for _ in x0], fprime=gradient,
                      fhess=hessian, fhess_p=hessian_p, disp=verbose, **args)
    return vector(RDF, min)


def minimize_constrained(func, cons, x0, gradient=None, algorithm='default', **args):
    r"""
    Minimize a function with constraints.

    INPUT:

    - ``func`` -- either a symbolic function, or a Python function whose
      argument is a tuple with `n` components

    - ``cons`` -- constraints. This should be either a function or list of
      functions that must be positive. Alternatively, the constraints can
      be specified as a list of intervals that define the region we are
      minimizing in. If the constraints are specified as functions, the
      functions should be functions of a tuple with `n` components
      (assuming `n` variables). If the constraints are specified as a list
      of intervals and there are no constraints for a given variable, that
      component can be (``None``, ``None``).

    - ``x0`` -- initial point for finding minimum

    - ``algorithm`` -- (optional) specify the algorithm to use:

      - ``'default'`` -- default choices

      - ``'l-bfgs-b'`` -- only effective if you specify bound constraints;
        see [ZBN1997]_

    - ``gradient`` -- (optional) gradient function. This will be computed
      automatically for symbolic functions. This is only used when the
      constraints are specified as a list of intervals.

    EXAMPLES:

    Let us maximize `x + y - 50` subject to the following constraints:
    `50x + 24y \leq 2400`, `30x + 33y \leq 2100`, `x \geq 45`,
    and `y \geq 5`::

        sage: f = lambda p: -p[0]-p[1]+50
        sage: c_1 = lambda p: p[0]-45
        sage: c_2 = lambda p: p[1]-5
        sage: c_3 = lambda p: -50*p[0]-24*p[1]+2400
        sage: c_4 = lambda p: -30*p[0]-33*p[1]+2100
        sage: a = minimize_constrained(f,[c_1,c_2,c_3,c_4],[2,3])
        sage: a
        (45.0, 6.25...)

    Let's find a minimum of `\sin(xy)`::

        sage: x,y = var('x y')                                                          # needs sage.symbolic
        sage: f(x,y) = sin(x*y)                                                         # needs sage.symbolic
        sage: minimize_constrained(f, [(None,None),(4,10)],[5,5])                       # needs sage.symbolic
        (4.8..., 4.8...)

    Check if L-BFGS-B finds the same minimum::

        sage: minimize_constrained(f, [(None,None),(4,10)],[5,5],                       # needs sage.symbolic
        ....:                      algorithm='l-bfgs-b')
        (4.7..., 4.9...)

    Rosenbrock function (see the :wikipedia:`Rosenbrock_function`)::

        sage: from scipy.optimize import rosen, rosen_der
        sage: minimize_constrained(rosen, [(-50,-10),(5,10)],[1,1],
        ....:                      gradient=rosen_der, algorithm='l-bfgs-b')
        (-10.0, 10.0)
        sage: minimize_constrained(rosen, [(-50,-10),(5,10)],[1,1],
        ....:                      algorithm='l-bfgs-b')
        (-10.0, 10.0)

    TESTS:

    Check if :issue:`6592` is fixed::

        sage: # needs sage.symbolic
        sage: x, y = var('x y')
        sage: f(x,y) = (100 - x) + (1000 - y)
        sage: c(x,y) = x + y - 479  # > 0
        sage: minimize_constrained(f, [c], [100, 300]) # random
        (805.985..., 1005.985...)
        sage: minimize_constrained(f, c, [100, 300])   # random
        (805.985..., 1005.985...)

    If ``func`` is symbolic, its minimizer should be in the same order
    as its arguments (:issue:`32511`)::

        sage: # needs sage.symbolic
        sage: x,y = SR.var('x,y')
        sage: f(y,x) = x - y
        sage: c1(y,x) = x
        sage: c2(y,x) = 1-y
        sage: minimize_constrained(f, [c1, c2], (0,0)) # abs tol 1e-04
        (1.0, 0.0)
    """
    from sage.structure.element import Expression
    from sage.ext.fast_callable import fast_callable
    import numpy
    if int(numpy.version.short_version[0]) > 1:
        numpy.set_printoptions(legacy="1.25")
    from scipy import optimize
    function_type = type(lambda x,y: x+y)

    if isinstance(func, Expression):
        var_list = func.arguments()
        fast_f = fast_callable(func, vars=var_list, domain=float)
        f = lambda p: fast_f(*p)
        gradient_list = func.gradient()
        fast_gradient_functions = [ fast_callable(gi,
                                                  vars=var_list,
                                                  domain=float)
                                    for gi in gradient_list ]
        gradient = lambda p: numpy.array([ a(*p) for a in fast_gradient_functions])
        if isinstance(cons, Expression):
            fast_cons = fast_callable(cons, vars=var_list, domain=float)
            cons = lambda p: numpy.array([fast_cons(*p)])
        elif isinstance(cons, list) and isinstance(cons[0], Expression):
            fast_cons = [ fast_callable(ci, vars=var_list, domain=float)
                          for ci in cons ]
            cons = lambda p: numpy.array([a(*p) for a in fast_cons])
    else:
        f = func

    if isinstance(cons, list):
        if isinstance(cons[0], (tuple, list)) or cons[0] is None:
            if gradient is not None:
                if algorithm == 'l-bfgs-b':
                    min = optimize.fmin_l_bfgs_b(f, x0, gradient, bounds=cons, **args)[0]
                else:
                    min = optimize.fmin_tnc(f, x0, gradient, bounds=cons, messages=0, **args)[0]
            else:
                if algorithm == 'l-bfgs-b':
                    min = optimize.fmin_l_bfgs_b(f, x0, approx_grad=True, bounds=cons, **args)[0]
                else:
                    min = optimize.fmin_tnc(f, x0, approx_grad=True, bounds=cons, messages=0, **args)[0]
        elif isinstance(cons[0], (function_type, Expression)):
            min = optimize.fmin_cobyla(f, x0, cons, **args)
    elif isinstance(cons, (function_type, Expression)):
        min = optimize.fmin_cobyla(f, x0, cons, **args)
    return vector(RDF, min)


def find_fit(data, model, initial_guess=None, parameters=None, variables=None, solution_dict=False):
    r"""
    Finds numerical estimates for the parameters of the function model to
    give a best fit to data.


    INPUT:

    - ``data`` -- a two dimensional table of floating point numbers of the
      form `[[x_{1,1}, x_{1,2}, \ldots, x_{1,k}, f_1],
      [x_{2,1}, x_{2,2}, \ldots, x_{2,k}, f_2],
      \ldots,
      [x_{n,1}, x_{n,2}, \ldots, x_{n,k}, f_n]]` given as either a list of
      lists, matrix, or numpy array.

    - ``model`` -- either a symbolic expression, symbolic function, or a
      Python function. ``model`` has to be a function of the variables
      `(x_1, x_2, \ldots, x_k)` and free parameters
      `(a_1, a_2, \ldots, a_l)`.

    - ``initial_guess`` -- (default: ``None``) initial estimate for the
      parameters `(a_1, a_2, \ldots, a_l)`, given as either a list, tuple,
      vector or numpy array. If ``None``, the default estimate for each
      parameter is `1`.

    - ``parameters`` -- (default: ``None``) a list of the parameters
      `(a_1, a_2, \ldots, a_l)`. If model is a symbolic function it is
      ignored, and the free parameters of the symbolic function are used.

    - ``variables`` -- (default: ``None``) a list of the variables
      `(x_1, x_2, \ldots, x_k)`. If model is a symbolic function it is
      ignored, and the variables of the symbolic function are used.

    - ``solution_dict`` -- boolean (default: ``False``); if ``True``, return the
      solution as a dictionary rather than an equation

    EXAMPLES:

    First we create some data points of a sine function with some "random"
    perturbations::

        sage: set_random_seed(0)
        sage: data = [(i, 1.2 * sin(0.5*i-0.2) + 0.1 * normalvariate(0, 1))             # needs sage.symbolic
        ....:         for i in xsrange(0, 4*pi, 0.2)]
        sage: var('a, b, c, x')                                                         # needs sage.symbolic
        (a, b, c, x)

    We define a function with free parameters `a`, `b` and `c`::

        sage: model(x) = a * sin(b * x - c)                                             # needs sage.symbolic

    We search for the parameters that give the best fit to the data::

        sage: find_fit(data, model)                                                     # needs sage.symbolic
        [a == 1.21..., b == 0.49..., c == 0.19...]

    We can also use a Python function for the model::

        sage: def f(x, a, b, c): return a * sin(b * x - c)
        sage: fit = find_fit(data, f, parameters=[a, b, c], variables=[x],              # needs sage.symbolic
        ....:                solution_dict = True)
        sage: fit[a], fit[b], fit[c]                                                    # needs sage.symbolic
        (1.21..., 0.49..., 0.19...)

    We search for a formula for the `n`-th prime number::

        sage: # needs sage.libs.pari
        sage: dataprime = [(i, nth_prime(i)) for i in range(1, 5000, 100)]
        sage: find_fit(dataprime, a * x * log(b * x),                                   # needs sage.symbolic
        ....:          parameters=[a, b], variables=[x])
        [a == 1.11..., b == 1.24...]


    ALGORITHM:

    Uses :func:`scipy:scipy.optimize.leastsq` which in turn uses MINPACK's
    ``lmdif`` and ``lmder`` algorithms.
    """
    import numpy
    if int(numpy.version.short_version[0]) > 1:
        numpy.set_printoptions(legacy="1.25")

    if not isinstance(data, numpy.ndarray):
        try:
            data = numpy.array(data, dtype=float)
        except (ValueError, TypeError):
            raise TypeError("data has to be a list of lists, a matrix, or a numpy array")
    elif data.dtype == object:
        raise ValueError("the entries of data have to be of type float")

    if data.ndim != 2:
        raise ValueError("data has to be a two dimensional table of floating point numbers")

    from sage.structure.element import Expression

    if isinstance(model, Expression):
        if variables is None:
            variables = list(model.arguments())
        if parameters is None:
            parameters = list(model.variables())
            for v in variables:
                parameters.remove(v)

    if data.shape[1] != len(variables) + 1:
        raise ValueError("each row of data needs %d entries, only %d entries given" % (len(variables) + 1, data.shape[1]))

    if parameters is None or len(parameters) == 0 or \
       variables is None or len(variables) == 0:
        raise ValueError("no variables given")

    if initial_guess is None:
        initial_guess = len(parameters) * [1]

    if not isinstance(initial_guess, numpy.ndarray):
        try:
            initial_guess = numpy.array(initial_guess, dtype=float)
        except (ValueError, TypeError):
            raise TypeError("initial_guess has to be a list, tuple, or numpy array")
    elif initial_guess.dtype == object:
        raise ValueError("the entries of initial_guess have to be of type float")

    if len(initial_guess) != len(parameters):
        raise ValueError("length of initial_guess does not coincide with the number of parameters")

    if isinstance(model, Expression):
        from sage.ext.fast_callable import fast_callable
        var_list = variables + parameters
        func = fast_callable(model, vars=var_list, domain=float)
    else:
        func = model

    def function(x_data, params):
        result = numpy.zeros(len(x_data))
        for row in range(len(x_data)):
            fparams = numpy.hstack((x_data[row], params)).tolist()
            result[row] = func(*fparams)
        return result

    def error_function(params, x_data, y_data):
        result = numpy.zeros(len(x_data))
        for row in range(len(x_data)):
            fparams = x_data[row].tolist() + params.tolist()
            result[row] = func(*fparams)
        return result - y_data

    x_data = data[:, 0:len(variables)]
    y_data = data[:, -1]

    from scipy.optimize import leastsq
    estimated_params, d = leastsq(error_function, initial_guess,
                                  args=(x_data, y_data))

    if isinstance(estimated_params, float):
        estimated_params = [estimated_params]
    else:
        estimated_params = estimated_params.tolist()

    if solution_dict:
        return dict(zip(parameters, estimated_params))

    return [item[0] == item[1] for item in zip(parameters, estimated_params)]


def binpacking(items, maximum=1, k=None, solver=None, verbose=0,
               *, integrality_tolerance=1e-3):
    r"""
    Solve the bin packing problem.

    The Bin Packing problem is the following :

    Given a list of items of weights `p_i` and a real value `k`, what is the
    least number of bins such that all the items can be packed in the bins,
    while ensuring that the sum of the weights of the items packed in each bin
    is at most `k` ?

    For more informations, see :wikipedia:`Bin_packing_problem`.

    Two versions of this problem are solved by this algorithm :

    - Is it possible to put the given items in `k` bins ?
    - What is the assignment of items using the least number of bins with
      the given list of items ?

    INPUT:

    - ``items`` -- list or dict; either a list of real values (the items'
      weight), or a dictionary associating to each item its weight

    - ``maximum`` -- (default: 1) the maximal size of a bin

    - ``k`` -- integer (default: ``None``); number of bins:

      - When set to an integer value, the function returns a partition of the
        items into `k` bins if possible, and raises an exception otherwise.

      - When set to ``None``, the function returns a partition of the items
        using the least possible number of bins.

    - ``solver`` -- (default: ``None``) specify a Mixed Integer Linear Programming
      (MILP) solver to be used. If set to ``None``, the default one is used. For
      more information on MILP solvers and which default solver is used, see
      the method
      :meth:`solve <sage.numerical.mip.MixedIntegerLinearProgram.solve>`
      of the class
      :class:`MixedIntegerLinearProgram <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- parameter for use with MILP solvers over an
      inexact base ring; see :meth:`MixedIntegerLinearProgram.get_values`

    OUTPUT:

    A list of lists, each member corresponding to a bin and containing either
    the list of the weights inside it when ``items`` is a list of items' weight,
    or the list of items inside it when ``items`` is a dictionary. If there is
    no solution, an exception is raised (this can only happen when ``k`` is
    specified or if ``maximum`` is less than the weight of one item).

    EXAMPLES:

    Trying to find the minimum amount of boxes for 5 items of weights
    `1/5, 1/4, 2/3, 3/4, 5/7`::

        sage: from sage.numerical.optimize import binpacking
        sage: values = [1/5, 1/3, 2/3, 3/4, 5/7]
        sage: bins = binpacking(values)                                                 # needs sage.numerical.mip
        sage: len(bins)                                                                 # needs sage.numerical.mip
        3

    Checking the bins are of correct size ::

        sage: all(sum(b) <= 1 for b in bins)                                            # needs sage.numerical.mip
        True

    Checking every item is in a bin ::

        sage: b1, b2, b3 = bins                                                         # needs sage.numerical.mip
        sage: all((v in b1 or v in b2 or v in b3) for v in values)                      # needs sage.numerical.mip
        True

    And only in one bin ::

        sage: sum(len(b) for b in bins) == len(values)                                  # needs sage.numerical.mip
        True

    One way to use only three boxes (which is best possible) is to put
    `1/5 + 3/4` together in a box, `1/3+2/3` in another, and `5/7`
    by itself in the third one.

    Of course, we can also check that there is no solution using only two boxes ::

        sage: from sage.numerical.optimize import binpacking
        sage: binpacking([0.2,0.3,0.8,0.9], k=2)                                        # needs sage.numerical.mip
        Traceback (most recent call last):
        ...
        ValueError: this problem has no solution

    We can also provide a dictionary keyed by items and associating to each item
    its weight. Then, the bins contain the name of the items inside it ::

        sage: values = {'a':1/5, 'b':1/3, 'c':2/3, 'd':3/4, 'e':5/7}
        sage: bins = binpacking(values)                                                 # needs sage.numerical.mip
        sage: set(flatten(bins)) == set(values.keys())                                  # needs sage.numerical.mip
        True

    TESTS:

    Wrong type for parameter items::

        sage: binpacking(set())
        Traceback (most recent call last):
        ...
        TypeError: parameter items must be a list or a dictionary.
    """
    if isinstance(items, list):
        weight = dict(enumerate(items))
    elif isinstance(items, dict):
        weight = items
    else:
        raise TypeError("parameter items must be a list or a dictionary.")

    if max(weight.values()) > maximum:
        raise ValueError("this problem has no solution")

    if k is None:
        from sage.functions.other import ceil
        k = ceil(sum(weight.values())/maximum)
        while True:
            from sage.numerical.mip import MIPSolverException
            try:
                return binpacking(items, k=k, maximum=maximum, solver=solver, verbose=verbose,
                                  integrality_tolerance=integrality_tolerance)
            except MIPSolverException:
                k = k + 1

    from sage.numerical.mip import MixedIntegerLinearProgram, MIPSolverException
    p = MixedIntegerLinearProgram(solver=solver)

    # Boolean variable indicating whether the ith element belongs to box b
    box = p.new_variable(binary=True)

    # Capacity constraint of each bin
    for b in range(k):
        p.add_constraint(p.sum(weight[i]*box[i,b] for i in weight) <= maximum)

    # Each item is assigned exactly one bin
    for i in weight:
        p.add_constraint(p.sum(box[i,b] for b in range(k)) == 1)

    try:
        p.solve(log=verbose)
    except MIPSolverException:
        raise ValueError("this problem has no solution")

    box = p.get_values(box, convert=bool, tolerance=integrality_tolerance)

    boxes = [[] for i in range(k)]

    for i,b in box:
        if box[i,b]:
            boxes[b].append(weight[i] if isinstance(items, list) else i)

    return boxes



from sage.data_structures.bitset cimport bitset_t

from sage.matroids.matroid cimport Matroid
from sage.matroids.basis_exchange_matroid cimport BasisExchangeMatroid
from sage.matroids.lean_matrix cimport LeanMatrix, GenericMatrix, BinaryMatrix, TernaryMatrix, QuaternaryMatrix


cdef class LinearMatroid(BasisExchangeMatroid):
    cdef LeanMatrix _A, _representation
    cdef long *_prow
    cdef object _zero, _one
    cdef dict _zonotopal_rho

    cpdef _forget(self)
    cpdef base_ring(self)
    cpdef characteristic(self)

    cdef list _setup_internal_representation(self, matrix, reduced_matrix, ring, keep_initial_representation)
    cdef _exchange_value_internal(self, long x, long y)

    cpdef representation(self, B=*, reduced=*, labels=*, order=*, lift_map=*)
    cpdef _current_rows_cols(self, B=*)
    cpdef representation_vectors(self)
    cpdef LeanMatrix _basic_representation(self, B=*)
    cpdef LeanMatrix _reduced_representation(self, B=*)

    cpdef bint _is_field_isomorphism(self, LinearMatroid other, morphism) noexcept
    cpdef is_field_equivalent(self, other)
    cpdef is_field_isomorphism(self, other, morphism)
    # cpdef is_field_isomorphic(self, other)  # TODO: currently only works as ``def``
    cpdef _fast_isom_test(self, other)
    cpdef relabel(self, mapping)

    cpdef _minor(self, contractions, deletions)
    cpdef dual(self)
    cpdef has_line_minor(self, k, hyperlines=*, certificate=*)
    cpdef has_field_minor(self, N)

    cpdef _exchange_value(self, e, f)
    cpdef fundamental_cycle(self, B, e)
    cpdef fundamental_cocycle(self, B, e)

    cpdef _line_ratios(self, F)
    cpdef _line_length(self, F)

    cpdef _line_cross_ratios(self, F)
    cpdef cross_ratios(self, hyperlines=*)
    cpdef cross_ratio(self, F, a, b, c, d)
    cpdef _line_cross_ratio_test(self, F, x, fundamentals)
    cpdef _cross_ratio_test(self, x, fundamentals, hyperlines=*)

    cpdef linear_extension(self, element, chain=*, col=*)
    cpdef linear_coextension(self, element, cochain=*, row=*)
    cpdef _linear_extensions(self, element, chains)
    cpdef _linear_coextensions(self, element, cochains)
    cdef _extend_chains(self, C, f, fundamentals=*)
    cpdef _linear_extension_chains(self, F, fundamentals=*)
    cpdef linear_extension_chains(self, F=*, simple=*, fundamentals=*)
    cpdef linear_coextension_cochains(self, F=*, cosimple=*, fundamentals=*)
    cpdef linear_extensions(self, element=*, F=*, simple=*, fundamentals=*)
    cpdef linear_coextensions(self, element=*, F=*, cosimple=*, fundamentals=*)

    cpdef _is_3connected_shifting(self, certificate=*)
    cpdef _is_4connected_shifting(self, certificate=*)

    cpdef is_valid(self, certificate=*)

    cpdef dict line_flats(self)
    cdef dict _zonotopal_rho_values(self)

cdef class BinaryMatroid(LinearMatroid):
    cdef tuple _b_invariant, _b_partition
    cdef BinaryMatrix _b_projection, _eq_part

    cpdef base_ring(self)
    cpdef characteristic(self)

    cpdef _current_rows_cols(self, B=*)
    cpdef LeanMatrix _basic_representation(self, B=*)
    cpdef LeanMatrix _reduced_representation(self, B=*)

    cdef  __fundamental_cocircuit(self, bitset_t, long x)

    cpdef _is_isomorphic(self, other, certificate=*)

    cpdef _minor(self, contractions, deletions)

    cpdef _make_invariant(self)
    cpdef _invariant(self)
    cpdef bicycle_dimension(self)
    cpdef brown_invariant(self)
    cpdef _principal_tripartition(self)
    cpdef BinaryMatrix _projection(self)
    cpdef BinaryMatrix _projection_partition(self)
    cpdef _fast_isom_test(self, other)
    cpdef relabel(self, mapping)

    cpdef bint is_graphic(self) noexcept
    cpdef is_valid(self, certificate=*)


cdef class TernaryMatroid(LinearMatroid):
    cdef object _two
    cdef tuple _t_invariant, _t_partition
    cdef TernaryMatrix _t_projection

    cpdef base_ring(self)
    cpdef characteristic(self)

    cpdef _current_rows_cols(self, B=*)
    cpdef LeanMatrix _basic_representation(self, B=*)
    cpdef LeanMatrix _reduced_representation(self, B=*)

    cdef  __fundamental_cocircuit(self, bitset_t, long x)

    cpdef _is_isomorphic(self, other, certificate=*)

    cpdef _minor(self, contractions, deletions)

    cpdef _make_invariant(self)
    cpdef _invariant(self)
    cpdef bicycle_dimension(self)
    cpdef character(self)
    cpdef _principal_quadripartition(self)
    cpdef TernaryMatrix _projection(self)
    cpdef _fast_isom_test(self, other)
    cpdef relabel(self, mapping)

    cpdef is_valid(self, certificate=*)

cdef class QuaternaryMatroid(LinearMatroid):
    cdef object _x_zero, _x_one
    cdef tuple _q_invariant, _q_partition
    cdef QuaternaryMatrix _q_projection

    cpdef base_ring(self)
    cpdef characteristic(self)

    cpdef _current_rows_cols(self, B=*)
    cpdef LeanMatrix _basic_representation(self, B=*)
    cpdef LeanMatrix _reduced_representation(self, B=*)

    cdef  __fundamental_cocircuit(self, bitset_t, long x)

    cpdef _is_isomorphic(self, other, certificate=*)

    cpdef _minor(self, contractions, deletions)

    cpdef _make_invariant(self)
    cpdef _invariant(self)
    cpdef bicycle_dimension(self)
    cpdef _principal_tripartition(self)
    cpdef _fast_isom_test(self, other)
    cpdef relabel(self, mapping)

    cpdef is_valid(self, certificate=*)

cdef class RegularMatroid(LinearMatroid):
    cdef _bases_count, _r_invariant
    cdef _r_projection, _r_hypergraph
    cdef _hypergraph_vertex_partition, _hypergraph_tuples

    cpdef base_ring(self)
    cpdef characteristic(self)

    cpdef _is_isomorphic(self, other, certificate=*)

    cpdef _invariant(self)
    cpdef _fast_isom_test(self, other)
    cpdef relabel(self, mapping)

    cpdef bases_count(self)
    cpdef _projection(self)
    cpdef _hypergraph(self)
    cdef _hypertest(self, other)
    cpdef has_line_minor(self, k, hyperlines=*, certificate=*)
    cpdef _linear_extension_chains(self, F, fundamentals=*)

    cpdef bint is_regular(self) noexcept
    cpdef bint is_graphic(self) noexcept
    cpdef is_valid(self, certificate=*)



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on, sig_off
from sage.ext.cplusplus cimport ccrepr

include 'misc.pxi'
include 'decl.pxi'

from sage.libs.ntl.ntl_ZZ cimport ntl_ZZ
from sage.libs.ntl.ntl_ZZX cimport ntl_ZZX
from cpython.object cimport PyObject_RichCompare

from sage.libs.ntl.ntl_ZZ import unpickle_class_args

cdef inline ntl_ZZ make_ZZ(ZZ_c* x):
    cdef ntl_ZZ y
    y = ntl_ZZ()
    y.x = x[0]
    del x
    return y

# You must do sig_on() before calling this function
cdef inline ntl_ZZ make_ZZ_sig_off(ZZ_c* x):
    cdef ntl_ZZ y = make_ZZ(x)
    sig_off()
    return y

cdef inline ntl_mat_ZZ make_mat_ZZ(mat_ZZ_c* x):
    cdef ntl_mat_ZZ y
    y = ntl_mat_ZZ.__new__(ntl_mat_ZZ)
    y.x = x[0]
    del x
    y.__nrows = y.x.NumRows()
    y.__ncols = y.x.NumCols()
    return y

# You must do sig_on() before calling this function
cdef inline ntl_mat_ZZ make_mat_ZZ_sig_off(mat_ZZ_c* x):
    cdef ntl_mat_ZZ y = make_mat_ZZ(x)
    sig_off()
    return y


##############################################################################
#
# ntl_mat_ZZ: Matrices over the integers via NTL
#
##############################################################################

cdef class ntl_mat_ZZ():
    # see ntl_mat_ZZ.pxd for data members
    r"""
    The \class{mat_ZZ} class implements arithmetic with matrices over `\Z`.
    """
    def __init__(self, nrows=0,  ncols=0, v=None):
        r"""
        The \class{mat_ZZ} class implements arithmetic with matrices over `\Z`.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(3,3) ; M
            [
            [0 0 0]
            [0 0 0]
            [0 0 0]
            ]
            sage: ntl.mat_ZZ(3,3,[1..9])
            [
            [1 2 3]
            [4 5 6]
            [7 8 9]
            ]
        """
        cdef unsigned long i, j
        cdef ntl_ZZ tmp
        if nrows == 0 and ncols == 0:
            return
        nrows = int(nrows)
        ncols = int(ncols)
        self.x.SetDims(nrows, ncols)
        self.__nrows = nrows
        self.__ncols = ncols
        if v is not None:
            for i from 0 <= i < nrows:
                for j from 0 <= j < ncols:
                    tmp = ntl_ZZ(v[i*ncols+j])
                    mat_ZZ_setitem(&self.x, i, j, &tmp.x)

    def __reduce__(self):
        """
        EXAMPLES::

            sage: m = ntl.mat_ZZ(3, 2, range(6)); m
            [
            [0 1]
            [2 3]
            [4 5]
            ]
            sage: loads(dumps(m))
            [
            [0 1]
            [2 3]
            [4 5]
            ]
            sage: loads(dumps(m)) == m
            True
        """
        return unpickle_class_args, (ntl_mat_ZZ, (self.__nrows, self.__ncols, self.list()))

    def __repr__(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,3,[5..10]) ; M.__repr__()
            '[\n[5 6 7]\n[8 9 10]\n]'
        """
        return ccrepr(self.x).replace('[[','[\n[',1)

    def __mul__(ntl_mat_ZZ self, other):
        """
        Multiply two matrices.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,2,[8..11]) ; N = ntl.mat_ZZ(2,2,[-1..2])
            sage: M*N
            [
            [1 18]
            [1 22]
            ]
        """
        cdef ntl_mat_ZZ r = ntl_mat_ZZ.__new__(ntl_mat_ZZ)
        if not isinstance(self, ntl_mat_ZZ):
            self = ntl_mat_ZZ(self)
        if not isinstance(other, ntl_mat_ZZ):
            other = ntl_mat_ZZ(other)
        sig_on()
        mat_ZZ_mul(r.x, (<ntl_mat_ZZ>self).x, (<ntl_mat_ZZ>other).x)
        sig_off()
        return r

    def __sub__(ntl_mat_ZZ self, other):
        """
        Return ``self - other``.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,2,[8..11]) ; N = ntl.mat_ZZ(2,2,[-1..2])
            sage: M-N
            [
            [9 9]
            [9 9]
            ]
        """
        cdef ntl_mat_ZZ r = ntl_mat_ZZ.__new__(ntl_mat_ZZ)
        if not isinstance(self, ntl_mat_ZZ):
            self = ntl_mat_ZZ(self)
        if not isinstance(other, ntl_mat_ZZ):
            other = ntl_mat_ZZ(other)
        sig_on()
        mat_ZZ_sub(r.x, (<ntl_mat_ZZ>self).x, (<ntl_mat_ZZ>other).x)
        sig_off()
        return r

    def __add__(ntl_mat_ZZ self, other):
        """
        Return ``self + other``.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,2,[8..11]) ; N = ntl.mat_ZZ(2,2,[-1..2])
            sage: M+N
            [
            [7 9]
            [11 13]
            ]
        """
        cdef ntl_mat_ZZ r = ntl_mat_ZZ.__new__(ntl_mat_ZZ)
        if not isinstance(self, ntl_mat_ZZ):
            self = ntl_mat_ZZ(self)
        if not isinstance(other, ntl_mat_ZZ):
            other = ntl_mat_ZZ(other)
        sig_on()
        mat_ZZ_add(r.x, (<ntl_mat_ZZ>self).x, (<ntl_mat_ZZ>other).x)
        sig_off()
        return r

    def __richcmp__(ntl_mat_ZZ self, other, int op):
        """
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,2,[3..6])
            sage: N = M^2
            sage: M == M
            True
            sage: M == N
            False
            sage: M == 0
            False
            sage: M < ntl.mat_ZZ(2,2,[4,4,4,4])
            True
            sage: M != ntl.mat_ZZ(1,1,[3])
            True
            sage: M != 0
            True
        """
        try:
            <ntl_mat_ZZ?>other
        except TypeError:
            return NotImplemented
        return PyObject_RichCompare(self.list(), other.list(), op)

    def __pow__(ntl_mat_ZZ self, long e, ignored):
        """
        Return ``self`` to the e power.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,2,[8..11])
            sage: M**4
            [
            [56206 62415]
            [69350 77011]
            ]
            sage: M**0
            [
            [1 0]
            [0 1]
            ]
            sage: M**(-1)
            Traceback (most recent call last):
            ...
            ValueError: cannot take negative powers of matrices.
        """
        if self.__nrows != self.__ncols:
            raise TypeError("cannot take powers of non-square matrices.")
        if e < 0:
            raise ValueError("cannot take negative powers of matrices.")
        cdef ntl_mat_ZZ r = ntl_mat_ZZ.__new__(ntl_mat_ZZ)
        sig_on()
        mat_ZZ_power(r.x, (<ntl_mat_ZZ>self).x, e)
        sig_off()
        return r

    def nrows(self):
        """
        Return the number of rows in ``self``.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(5,5,range(25))
            sage: M.nrows()
            5
        """
        return self.__nrows

    def ncols(self):
        """
        Return the number of columns in ``self``.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(5,8,range(40))
            sage: M.ncols()
            8
        """
        return self.__ncols

    def __setitem__(self, ij, x):
        """
        Given a tuple (i, j), return self[i,j].

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,9,[3..20])
            sage: M[1,7] ## indirect doctest
            19
        """
        cdef ntl_ZZ y
        cdef int i, j
        if not isinstance(x, ntl_ZZ):
            y = ntl_ZZ(x)
        else:
            y = x
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError('ij must be a 2-tuple')
        i, j = int(ij[0]),int(ij[1])
        if i < 0 or i >= self.__nrows or j < 0 or j >= self.__ncols:
            raise IndexError("array index out of range")
        sig_on()
        mat_ZZ_setitem(&self.x, i, j, &y.x)
        sig_off()

    def __getitem__(self, ij):
        """
        EXAMPLES::

            sage: m = ntl.mat_ZZ(3, 2, range(6))
            sage: m[0,0] ## indirect doctest
            0
            sage: m[2,1]
            5
            sage: m[3,2] #  oops, 0 based
            Traceback (most recent call last):
            ...
            IndexError: array index out of range
        """
        cdef int i, j
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError('ij must be a 2-tuple')
        i, j = ij
        if i < 0 or i >= self.__nrows or j < 0 or j >= self.__ncols:
            raise IndexError("array index out of range")
        sig_on()
        return make_ZZ_sig_off(mat_ZZ_getitem(&self.x, i+1, j+1))

    def list(self):
        """
        EXAMPLES::

            sage: m = ntl.mat_ZZ(3, 4, range(12)); m
            [
            [0 1 2 3]
            [4 5 6 7]
            [8 9 10 11]
            ]
            sage: m.list()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        """
        cdef int i, j
        sig_on()
        L = [make_ZZ(mat_ZZ_getitem(&self.x, i+1, j+1))
                    for i from 0 <= i < self.__nrows
                        for j from 0 <= j < self.__ncols]
        sig_off()
        return L

    def determinant(self, deterministic=True):
        """
        Return the determinant of ``self``.

        EXAMPLES::

            sage: ntl.mat_ZZ(8,8,[3..66]).determinant()
            0
            sage: ntl.mat_ZZ(2,5,range(10)).determinant()
            Traceback (most recent call last):
            ...
            TypeError: cannot take determinant of non-square matrix.
            sage: ntl.mat_ZZ(4,4,[next_prime(2**i) for i in range(16)]).determinant()
            -10248
            sage: ntl.mat_ZZ(4,4,[ ZZ.random_element() for _ in range(16) ]).determinant()  # random
            678
        """
        if self.__nrows != self.__ncols:
            raise TypeError("cannot take determinant of non-square matrix.")
        sig_on()
        return make_ZZ_sig_off(mat_ZZ_determinant(&self.x, deterministic))

    def HNF(self, D=None):
        r"""
        The input matrix A=self is an n x m matrix of rank m (so n >=
        m), and D is a multiple of the determinant of the lattice L
        spanned by the rows of A.  The output W is the Hermite Normal
        Form of A; that is, W is the unique m x m matrix whose rows
        span L, such that

        - W is lower triangular,
        - the diagonal entries are positive,
        - any entry below the diagonal is a nonnegative number
          strictly less than the diagonal entry in its column.

        This is implemented using the algorithm of [P. Domich,
        R. Kannan and L. Trotter, Math. Oper. Research 12:50-59,
        1987].

        TIMINGS:

        NTL is not very good compared to MAGMA, unfortunately::

            sage: a = MatrixSpace(ZZ,200).random_element(x=-2, y=2)    # -2 to 2
            sage: A = ntl.mat_ZZ(200,200)
            sage: for i in range(a.nrows()):
            ....:     for j in range(a.ncols()):
            ....:         A[i,j] = a[i,j]
            sage: t = cputime(); d = A.determinant()
            sage: cputime(t)          # random
            0.33201999999999998
            sage: t = cputime(); B = A.HNF(d)  # long time (5s on sage.math, 2011)
            sage: cputime(t)          # random
            6.4924050000000006

        In comparison, MAGMA does this much more quickly:
        \begin{verbatim}
            > A := MatrixAlgebra(IntegerRing(),200)![Random(-2,2) : i in [1..200^2]];
            > time d := Determinant(A);
            Time: 0.140
            > time H := HermiteForm(A);
            Time: 0.290
        \end{verbatim}

        Also, PARI is also faster than NTL if one uses the flag 1 to
        the mathnf routine.  The above takes 16 seconds in PARI.

        TESTS::

            sage: ntl.mat_ZZ(2,2,[1..4]).HNF()
            [
            [1 0]
            [0 2]
            ]
        """
        cdef ntl_ZZ _D
        if D is None:
            _D = self.determinant()
        else:
            _D = ntl_ZZ(D)
        sig_on()
        return make_mat_ZZ_sig_off(mat_ZZ_HNF(&self.x, &_D.x))

    def charpoly(self):
        """
        Find the characteristic polynomial of self, and return it
        as an NTL ZZX.

        EXAMPLES::

            sage: M = ntl.mat_ZZ(2,2,[1,2,3,4])
            sage: M.charpoly()
            [-2 -5 1]
            sage: type(_)
            <class 'sage.libs.ntl.ntl_ZZX.ntl_ZZX'>
            sage: M.determinant()
            -2
        """
        cdef ntl_ZZX r = ntl_ZZX()
        sig_on()
        mat_ZZ_CharPoly(r.x, self.x)
        sig_off()
        return r

    def BKZ_FP(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.BKZ_FP(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.BKZ_FP(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_BKZ_FP(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_BKZ_FP_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def BKZ_QP(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.BKZ_QP(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.BKZ_QP(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_BKZ_QP(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_BKZ_QP_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def BKZ_QP1(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.BKZ_QP1(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.BKZ_QP1(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_BKZ_QP1(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_BKZ_QP1_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def BKZ_XD(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.BKZ_XD(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.BKZ_XD(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_BKZ_XD(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_BKZ_XD_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def BKZ_RR(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.BKZ_RR(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.BKZ_RR(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_BKZ_RR(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_BKZ_RR_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def G_BKZ_FP(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.G_BKZ_FP(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.G_BKZ_FP(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_G_BKZ_FP(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_G_BKZ_FP_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def G_BKZ_QP(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.G_BKZ_QP(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.G_BKZ_QP(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_G_BKZ_QP(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_G_BKZ_QP_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def G_BKZ_QP1(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.G_BKZ_QP1(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.G_BKZ_QP1(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_G_BKZ_QP1(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_G_BKZ_QP1_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def G_BKZ_XD(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.G_BKZ_XD(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.G_BKZ_XD(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_G_BKZ_XD(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_G_BKZ_XD_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def G_BKZ_RR(self, U=None, delta=0.99, BlockSize=10, prune=0, verbose=False):
        r"""
        All BKZ methods are equivalent to the LLL routines,
        except that Block Korkin-Zolotarev reduction is applied. We
        describe here only the differences in the calling syntax.

        * The optional parameter "BlockSize" specifies the size of the
        blocks in the reduction. High values yield shorter vectors,
        but the running time increases exponentially with BlockSize.
        BlockSize should be between 2 and the number of rows of B.

        * The optional parameter "prune" can be set to any positive
        number to invoke the Volume Heuristic from [Schnorr and
        Horner, Eurocrypt '95].  This can significantly reduce the
        running time, and hence allow much bigger block size, but the
        quality of the reduction is of course not as good in general.
        Higher values of prune mean better quality, and slower running
        time.  When prune == 0, pruning is disabled.  Recommended
        usage: for BlockSize >= 30, set 10 <= prune <= 15.

        * The QP1 variant uses quad_float precision to compute
        Gram-Schmidt, but uses double precision in the search phase
        of the block reduction algorithm.  This seems adequate for
        most purposes, and is faster than QP, which uses quad_float
        precision uniformly throughout.

        INPUT:

        - ``U`` -- permutation matrix (see LLL, default: ``None``)
        - ``delta`` -- reduction parameter (default: 0.99)
        - ``BlockSize`` -- see above (default: 10)
        - ``prune`` -- see above (default: 0)
        - ``verbose`` -- print verbose output (default: ``False``)

        EXAMPLES::

            sage: A = Matrix(ZZ,5,5,range(25))
            sage: a = A._ntl_()
            sage: a.G_BKZ_RR(); a
            2
            [
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 1 2 3 4]
            [5 3 1 -1 -3]
            ]

            sage: U = ntl.mat_ZZ(2,2) # note that the dimension doesn't matter
            sage: r = a.G_BKZ_RR(U=U); U
            [
            [0 1 0 0 0]
            [1 0 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]
            ]
        """
        if U is None:
            sig_on()
            rank = mat_ZZ_G_BKZ_RR(self.x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        elif isinstance(U, ntl_mat_ZZ):
            sig_on()
            rank = mat_ZZ_G_BKZ_RR_U(self.x, (<ntl_mat_ZZ>U).x, float(delta), int(BlockSize), int(prune), 0, int(verbose))
            sig_off()
            return rank
        else:
            raise TypeError("parameter U has wrong type.")

    def LLL(self, a=3, b=4, return_U=False, verbose=False):
        r"""
        Perform LLL reduction of ``self`` (puts \code{self} in an LLL form).

        \code{self} is an `m x n` matrix, viewed as `m` rows of
        `n`-vectors.  `m` may be less than, equal to, or greater than `n`,
        and the rows need not be linearly independent. ``self`` is
        transformed into an LLL-reduced basis, and the return value is
        the rank r of ``self`` so as det2 (see below).  The first `m-r` rows
        of ``self`` are zero.

        More specifically, elementary row transformations are
        performed on \code{self} so that the nonzero rows of
        new-\code{self} form an LLL-reduced basis for the lattice
        spanned by the rows of old-\code{self}.  The default reduction
        parameter is `\delta=3/4`, which means that the squared length
        of the first nonzero basis vector is no more than `2^{r-1}`
        times that of the shortest vector in the lattice.

        det2 is calculated as the \emph{square} of the determinant of
        the lattice---note that sqrt(det2) is in general an integer
        only when r = n.

        If return_U is True, a value U is returned which is the
        transformation matrix, so that U is a unimodular m x m matrix
        with U * old-\code{self} = new-\code{self}. Note that the
        first m-r rows of U form a basis (as a lattice) for the kernel
        of old-B.

        The parameters a and b allow an arbitrary reduction parameter
        `\delta=a/b`, where `1/4 < a/b \leq 1`, where a and b are positive
        integers.  For a basis reduced with parameter delta, the
        squared length of the first nonzero basis vector is no more
        than `1/(delta-1/4)^{r-1}` times that of the shortest vector in
        the lattice.

        The algorithm employed here is essentially the one in Cohen's
        book: [H. Cohen, A Course in Computational Algebraic Number
        Theory, Springer, 1993]

        INPUT:

        - ``a`` -- parameter a as described above (default: 3)
        - ``b`` -- parameter b as described above (default: 4)
        - ``return_U`` -- return U as described above
        - ``verbose`` -- if ``True`` NTL will produce some verbatim messages on
          what's going on internally (default: ``False``)

        OUTPUT:

        (rank,det2,[U]) where rank,det2, and U are as described
        above and U is an optional return value if return_U is ``True``.

        EXAMPLES::

            sage: M=ntl.mat_ZZ(3,3,[1,2,3,4,5,6,7,8,9])
            sage: M.LLL()
            (2, 54)
            sage: M
            [
            [0 0 0]
            [2 1 0]
            [-1 1 3]
            ]
            sage: M=ntl.mat_ZZ(4,4,[-6,9,-15,-18,4,-6,10,12,10,-16,18,35,-24,36,-46,-82]); M
            [
            [-6 9 -15 -18]
            [4 -6 10 12]
            [10 -16 18 35]
            [-24 36 -46 -82]
            ]
            sage: M.LLL()
            (3, 19140)
            sage: M
            [
            [0 0 0 0]
            [0 -2 0 0]
            [-2 1 -5 -6]
            [0 -1 -7 5]
            ]

        WARNING: This method modifies \code{self}. So after applying
        this method your matrix will be a vector of vectors.
        """
        cdef ZZ_c *det2
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_LLL_U(&det2, &self.x, &U.x, int(a), int(b), int(verbose)))

            return rank, make_ZZ_sig_off(det2), U
        else:
            sig_on()
            rank = int(mat_ZZ_LLL(&det2,&self.x,int(a),int(b),int(verbose)))
            return rank, make_ZZ_sig_off(det2)

    def LLL_FP(self, delta=0.75, return_U=False, verbose=False):
        r"""
        Perform approximate LLL reduction of \code{self} (puts
        \code{self} in an LLL form) subject to the following
        conditions:

        The precision is double.

        The return value is the rank of B.

        Classical Gram-Schmidt Orthogonalization is used:

        This choice uses classical methods for computing the
        Gram-Schmidt orthogonalization.  It is fast but prone to
        stability problems.  This strategy was first proposed by
        Schnorr and Euchner [C. P. Schnorr and M. Euchner,
        Proc. Fundamentals of Computation Theory, LNCS 529, pp. 68-85,
        1991].  The version implemented here is substantially
        different, improving both stability and performance.

        If return_U is True, then also U is returned which is
        the transition matrix: `U * self_{old} = self_{new}`

        The optional argument 'delta' is the reduction parameter, and
        may be set so that 0.50 <= delta < 1.  Setting it close to 1
        yields shorter vectors, and also improves the stability, but
        increases the running time.  Recommended value: delta =
        0.99.

        The optional parameter 'verbose' can be set to see all kinds
        of fun things printed while the routine is executing.  A
        status report is also printed every once in a while.

        INPUT:

        - ``delta`` -- as described above (0.5 <= delta < 1.0) (default: 0.75)
        - ``return_U`` -- return U as described above
        - ``verbose`` -- if ``True`` NTL will produce some verbatim messages on
          what's going on internally (default: ``False``)

        OUTPUT:

        (rank,[U]) where rank and U are as described above and U
        is an optional return value if ``return_U`` is ``True``.

        EXAMPLES::

            sage: M=ntl.mat_ZZ(3,3,[1,2,3,4,5,6,7,8,9])
            sage: M.LLL_FP()
            2
            sage: M
            [
            [0 0 0]
            [2 1 0]
            [-1 1 3]
            ]
            sage: M=ntl.mat_ZZ(4,4,[-6,9,-15,-18,4,-6,10,12,10,-16,18,35,-24,36,-46,-82]); M
            [
            [-6 9 -15 -18]
            [4 -6 10 12]
            [10 -16 18 35]
            [-24 36 -46 -82]
            ]
            sage: M.LLL_FP()
            3
            sage: M
            [
            [0 0 0 0]
            [0 -2 0 0]
            [-2 1 -5 -6]
            [0 -1 -7 5]
            ]

        WARNING: This method modifies \code{self}. So after applying this
        method your matrix will be a vector of vectors.
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_LLL_FP_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_LLL_FP(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    def LLL_QP(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as \code{self.LLL_FP} using the
        same calling conventions but with quad float precision.

        EXAMPLES::

            sage: M=ntl.mat_ZZ(3,3,[1,2,3,4,5,6,7,8,9])
            sage: M.LLL_QP(delta=0.75)
            2
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_LLL_QP_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_LLL_QP(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    def LLL_XD(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as \code{self.LLL_FP} using the
        same calling conventions but with extended exponent double
        precision.

        EXAMPLES::

            sage: M=ntl.mat_ZZ(3,3,[1,2,3,4,5,6,7,8,9])
            sage: M.LLL_XD(delta=0.75)
            2
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_LLL_XD_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_LLL_XD(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    def LLL_RR(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as \code{self.LLL_FP} using the
        same calling conventions but with arbitrary precision floating
        point numbers.

        EXAMPLES::

            sage: M=ntl.mat_ZZ(3,3,[1,2,3,4,5,6,7,8,9])
            sage: M.LLL_RR(delta=0.75)
            2
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ.__new__(ntl_mat_ZZ)
            sig_on()
            rank = int(mat_ZZ_LLL_RR_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_LLL_RR(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    # Givens Orthogonalization.  This is a bit slower, but generally
    # much more stable, and is really the preferred orthogonalization
    # strategy.  For a nice description of this, see Chapter 5 of
    # [G. Golub and C. van Loan, Matrix Computations, 3rd edition,
    # Johns Hopkins Univ. Press, 1996].

    def G_LLL_FP(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as self.LLL_FP using the same
        calling conventions but uses the Givens Orthogonalization.

        Givens Orthogonalization.  This is a bit slower, but generally
        much more stable, and is really the preferred
        orthogonalization strategy.  For a nice description of this,
        see Chapter 5 of [G. Golub and C. van Loan, Matrix
        Computations, 3rd edition, Johns Hopkins Univ. Press, 1996].
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_G_LLL_FP_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_G_LLL_FP(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    def G_LLL_QP(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as self.G_LLL_FP using the same
        calling conventions but with quad float precision.
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_G_LLL_QP_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_G_LLL_QP(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    def G_LLL_XD(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as self.G_LLL_FP using the same
        calling conventions but with extended exponent double
        precision.
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_G_LLL_XD_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_G_LLL_XD(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank

    def G_LLL_RR(self, delta, return_U=False, verbose=False):
        r"""
        Perform the same reduction as self.G_LLL_FP using the same
        calling conventions but with arbitrary precision floating
        point numbers.
        """
        cdef ntl_mat_ZZ U
        if return_U:
            U = ntl_mat_ZZ(self.__nrows, self.__nrows)
            sig_on()
            rank = int(mat_ZZ_G_LLL_RR_U(self.x, U.x, float(delta), 0, 0, int(verbose)))
            sig_off()
            return rank, U
        else:
            sig_on()
            rank = int(mat_ZZ_G_LLL_RR(self.x,float(delta),0,0,int(verbose)))
            sig_off()
            return rank



from sage.cli.eval_cmd import EvalCmd
from sage.cli.options import CliOptions
import pytest


def test_eval_cmd_print(capsys):
    options = CliOptions(command="print(3^33)")
    eval_cmd = EvalCmd(options)

    result = eval_cmd.run()
    captured = capsys.readouterr()
    assert captured.out == "5559060566555523\n"
    assert result == 0


def test_eval_cmd_invalid_command(capsys):
    options = CliOptions(command="invalid_command")
    eval_cmd = EvalCmd(options)

    with pytest.raises(NameError) as err:
        result = eval_cmd.run()
    assert str(err.value) == "name 'invalid_command' is not defined"
